# -*- coding: utf-8 -*-
"""Synteny hash operations."""


# third-party imports
import attr
import numpy as np
import pandas as pd

# module imports
from .common import DIRECTIONAL_CATEGORY
from .common import hash_array
from .common import logger

# helper functions


def _cum_val_count(arr):
    """Return an array of cumulative counts of values."""
    counts = {}
    out_arr = pd.array(
        [pd.NA] * len(arr),
        dtype=pd.UInt32Dtype(),
    )
    for i, val in enumerate(arr):
        if pd.isnull(val):
            continue
        elif val in counts:
            counts[val] += 1
        else:
            counts[val] = 1
        out_arr[i] = counts[val]
    return out_arr


def _true_positions_and_runs(bool_vec):
    """Return arrays of positions and lengths of runs of True."""
    uneq_idxs = np.append(
        np.where(bool_vec[1:] != bool_vec[:-1]), bool_vec.size - 1
    )
    runlengths = np.diff(np.append(-1, uneq_idxs))
    positions = np.cumsum(np.append(0, runlengths))[:-1]
    true_idxs = np.where(bool_vec[positions])
    return positions[true_idxs], runlengths[true_idxs]


def _fill_na_with_last_valid(ser, flip=False):
    """Input a series with NA values, returns a series with those values filled."""
    lv_arr = pd.array(
        [pd.NA] * len(ser),
        dtype=pd.UInt32Dtype(),
    )
    if not (ser.isnull().all() or ser.notna().all()):
        null_vec = ser.isnull().to_numpy()
        val_vec = ser.to_numpy()
        if flip:
            null_vec = np.flip(null_vec)
            val_vec = np.flip(val_vec)
        first_null_pos, null_runs = _true_positions_and_runs(null_vec)
        fill_vals = np.append(pd.NA, val_vec)[first_null_pos]
        for i, pos in enumerate(first_null_pos):
            for j in range(null_runs[i]):
                lv_arr[pos + j] = fill_vals[i]
        if flip:
            lv_arr = np.flip(lv_arr)
        lv_ser = pd.Series(lv_arr, index=ser.index)
        return lv_ser


def _cum_val_cnt_where_ser2_is_na(ser1, ser2, flip=False):
    """Return the cumulative value count of ser1 in regions where ser2 is NA."""
    if len(ser1) != len(ser2):
        logger.warning(f"Lengths of ser1 and ser2 differ at {ser1}")
    vc_arr = pd.array(
        [pd.NA] * len(ser1),
        dtype=pd.UInt32Dtype(),
    )
    if not (ser2.isnull().all() or ser2.notna().all()):
        null_vec = ser2.isnull().to_numpy()
        val_vec = ser1.to_numpy()
        if flip:
            null_vec = np.flip(null_vec)
            val_vec = np.flip(val_vec)
        null_pos, null_runs = _true_positions_and_runs(null_vec)
        null_len = len(null_pos)
        for i in range(null_len):
            vc_arr[
                null_pos[i] : (null_pos[i] + null_runs[i])
            ] = _cum_val_count(
                val_vec[null_pos[i] : (null_pos[i] + null_runs[i])]
            )
        if flip:
            vc_arr = np.flip(vc_arr)
    vc_ser = pd.Series(vc_arr, index=ser2.index)
    return vc_ser


@attr.s
class SyntenyBlockHasher(object):
    """Synteny-block hashes via reversible-peatmer method."""

    k = attr.ib(default=3)
    peatmer = attr.ib(default=True)
    thorny = attr.ib(default=True)
    disambig_adj_only = attr.ib(default=True)
    prefix = attr.ib(default="syn")

    def hash_name(self, no_prefix=False):
        """Return the string name of the hash function."""
        if no_prefix:
            prefix_str = ""
        else:
            prefix_str = self.prefix + "."
        if self.thorny:
            thorny_str = "thorny"
        else:
            thorny_str = ""
        if self.peatmer:
            return f"{prefix_str}hash.{thorny_str}peatmer{self.k}"
        return f"{prefix_str}hash.{thorny_str}kmer{self.k}"

    def shingle(self, cluster_series, direction, hash_val):
        """Return a vector of anchor ID's. """
        vec = cluster_series.to_numpy().astype(int)
        steps = np.insert((vec[1:] != vec[:-1]).astype(int), 0, 0).cumsum()
        if max(steps) != self.k - 1:
            logger.warning(f"Inconsistency in shingling hash {hash_val}")
            logger.warning(f"input homology string={vec}")
            steps[np.where(steps > self.k - 1)] = self.k - 1
        if direction == "-":
            steps = self.k - 1 - steps
        return steps

    def calculate(self, cluster_series):
        """Return an array of synteny block hashes data."""
        # Maybe the best code I've ever written--JB
        vec = cluster_series.to_numpy().astype(int)
        if self.peatmer:
            uneq_idxs = np.append(np.where(vec[1:] != vec[:-1]), vec.size - 1)
            runlengths = np.diff(np.append(-1, uneq_idxs))
            positions = np.cumsum(np.append(0, runlengths))[:-1]
            n_mers = len(positions) - self.k + 1
            footprints = pd.array(
                [runlengths[i : i + self.k].sum() for i in range(n_mers)],
                dtype=pd.UInt32Dtype(),
            )
        else:
            n_elements = len(cluster_series)
            n_mers = n_elements - self.k + 1
            positions = np.arange(n_elements)
            footprints = pd.array([self.k] * n_mers, dtype=pd.UInt32Dtype())
        if n_mers < 1:
            return None
        # Calculate k-mers over indirect index
        kmer_mat = np.array(
            [vec[positions[i : i + self.k]] for i in range(n_mers)]
        )
        fwd_rev_hashes = np.array(
            [
                np.apply_along_axis(hash_array, 1, kmer_mat),
                np.apply_along_axis(hash_array, 1, np.flip(kmer_mat, axis=1)),
            ]
        )
        plus_minus = np.array([["+"] * n_mers, ["-"] * n_mers])
        directions = np.take_along_axis(
            plus_minus,
            np.expand_dims(fwd_rev_hashes.argmin(axis=0), axis=0),
            axis=0,
        )[0]
        return pd.DataFrame(
            [
                pd.Categorical(directions, dtype=DIRECTIONAL_CATEGORY),
                footprints,
                pd.array(
                    np.amin(fwd_rev_hashes, axis=0), dtype=pd.UInt32Dtype()
                ),
            ],
            columns=[
                "syn.hash.direction",
                "syn.hash.footprint",
                self.hash_name(),
            ],
            index=cluster_series.index[positions[:n_mers]],
        )

    def calculate_disambig_hashes(self, df):
        """Calculate disambiguation frame (per-fragment).

        if self.disambig_adj_only is True, then disambiguation will be done
        only for those locations adjacent to an umabiguous hash.
        """
        hash2_fr = df[["syn.anchor.id", "tmp.ambig.id"]].copy()
        hash2_fr = hash2_fr.rename(columns={"syn.anchor.id": "tmp.anchor.id"})
        hash2_fr["tmp.upstr_anchor"] = _fill_na_with_last_valid(
            df["syn.anchor.id"]
        )
        hash2_fr["tmp.downstr_anchor"] = _fill_na_with_last_valid(
            df["syn.anchor.id"], flip=True
        )
        hash2_fr["tmp.upstr_occur"] = _cum_val_cnt_where_ser2_is_na(
            df["tmp.ambig.id"], df["syn.anchor.id"]
        )
        hash2_fr["tmp.downstr_occur"] = _cum_val_cnt_where_ser2_is_na(
            df["tmp.ambig.id"], df["syn.anchor.id"], flip=True
        )
        hash2_fr["tmp.i"] = range(len(hash2_fr))
        upstream_hash = pd.array(
            [pd.NA] * len(hash2_fr), dtype=pd.UInt32Dtype()
        )
        downstream_hash = pd.array(
            [pd.NA] * len(hash2_fr), dtype=pd.UInt32Dtype()
        )
        hash2_fr["tmp.disambig.up"] = pd.NA
        hash2_fr["tmp.disambig.down"] = pd.NA
        for unused_id, row in hash2_fr.iterrows():
            row_no = row["tmp.i"]
            ambig_base = row["tmp.ambig.id"]
            upstream_unambig = row["tmp.upstr_anchor"]
            downstream_unambig = row["tmp.downstr_anchor"]
            occur_upstream = row["tmp.upstr_occur"]
            occur_downstream = row["tmp.downstr_occur"]
            if pd.notna(ambig_base):
                if pd.notna(upstream_unambig):
                    if not pd.notna(occur_upstream):
                        logger.warning(
                            f"Something is wrong upstream of base {ambig_base}"
                        )
                    if self.disambig_adj_only and occur_upstream > 1:
                        continue
                    upstream_hash[row_no] = hash_array(
                        np.array(
                            [upstream_unambig, ambig_base, occur_upstream]
                        )
                    )
                if pd.notna(downstream_unambig):
                    if not pd.notna(occur_downstream):
                        logger.warning(
                            f"Something is wrong downstream of base {ambig_base}"
                        )
                    if self.disambig_adj_only and occur_downstream > 1:
                        continue
                    downstream_hash[row_no] = hash_array(
                        np.array(
                            [ambig_base, downstream_unambig, occur_downstream]
                        )
                    )
        hash2_fr["tmp.disambig.up"] = upstream_hash
        hash2_fr["tmp.disambig.down"] = downstream_hash
        return hash2_fr[["tmp.disambig.up", "tmp.disambig.down"]]



# -*- coding: utf-8 -*-
"""Constants and functions in common across modules."""
# standard library imports
import contextlib
import mmap
import os
import sys
import tempfile
from pathlib import Path

# third-party imports
import numpy as np
import pandas as pd
import xxhash
from loguru import logger as loguru_logger
from memory_tempfile import MemoryTempfile


# global constants
NAME = "azulejo"
DEFAULT_PARQUET_COMPRESSION = "ZSTD"
# Sizes and minimum read times with various compressions
# for a file with one proteome on a system with M.2 SSD disk
# under pyarrow 1.0.0 into pandas 1.1.0:
#   "NONE": 43MB, 1.8s
#   "ZSTD": 13M, 1.8s
# "SNAPPY": 29 MB, 1.8s
# "BROTLI": 13 MB, 1.9s
#    "LZ4": 23MB, (disabled under pyarrow 1.0.0, was about like brotli under 0.17)
#   "GZIP": 14 MB, 2.1 s
#    "LZO": not supported
#    "BZ2": not supported
# In addition, the ingest process took 28.8s with None, and
# 28.4 s with ZSTD, probably due to writing less data.
# With its 70% compression factor,  ZSTD can be expected to
# perform even better relative to uncompressed and snappy
# on production systems with slower disks for which
# cache is not warmed up (as mine was in this test).
# So ZSTD seems a clear choice for now.

PARQUET_EXTENSIONS = ["parquet", "pq", "parq"]
TSV_EXTENSIONS = ["tsv"]
SAVED_INPUT_FILE = "input.toml"

# Changing the extension of these files will change the type of file written.
# TSV files, though readable/editable, do not give the written values back.
# Parquet is also ~100X faster.

CLUSTER_FILETYPE = "parq"
CLUSTERS_FILE = "homology_clusters.parq"
CLUSTERSYN_FILE = "homology_clusters.syn.parq"
CLUSTER_HIST_FILE = "homology_cluster_hist.tsv"
FRAGMENTS_FILE = "fragments.tsv"
ANCHOR_HIST_FILE = "anchor_hist.tsv"
HOMOLOGY_FILE = "proteins.hom.parq"
PROTEOMES_FILE = "proteomes.tsv"
PROTEOMOLOGY_FILE = "proteomes.hom.parq"
PROTEOSYN_FILE = "proteomes.hom.syn.parq"
PROTEINS_FILE = "proteins.parq"
SYNTENY_FILE = "proteins.hom.syn.parq"
ANCHORS_FILE = "synteny_anchors.tsv"
SYNTENY_FILETYPE = "tsv"
COLLECTION_FILE = "collection.json"
COLLECTION_HOM_FILE = "collection.hom.json"
COLLECTION_SYN_FILE = "collection.hom.syn.json"
EXTERNAL_CLUSTERS_FILE = "homology_clusters-external.tsv"

# fragment-name defs
PLASTID_STARTS = ["chromop", "chl", "mt", "mi", "rh", "mu", "le", "pl"]
CHROMOSOME_SYNONYMS = ["chromosome", "chrom", "chro", "gs", "gm"]
ALTERNATE_ABBREV = "alt"
CHROMOSOME_ABBREV = "chr"
SCAFFOLD_SYNONYMS = ["scaffold", "scaf", "sca"]
SCAFFOLD_ABBREV = "sc"

# synteny codes
UNAMBIGUOUS_CODE = "U"
DISAMBIGUATED_CODE = "D"
INDIRECT_CODE = "I"
LOCALLY_UNAMBIGUOUS_CODE = "L"
NON_AMBIGUOUS_CODE = "N"
AMBIGUOUS_CODE = "A"
CODE_DICT = {
    UNAMBIGUOUS_CODE: "unambiguous",
    DISAMBIGUATED_CODE: "disambiguated",
    INDIRECT_CODE: "indirectly unambiguous",
    LOCALLY_UNAMBIGUOUS_CODE: "locally unambiguous",
    NON_AMBIGUOUS_CODE: "non-ambiguous",
    AMBIGUOUS_CODE: "ambiguous",
}
DIRECTIONAL_CATEGORY = pd.CategoricalDtype(categories=["-", "+"])
YES_NO = pd.CategoricalDtype(categories=["y", "n"])
SYNTENY_CATEGORY = pd.CategoricalDtype(categories=CODE_DICT.keys())
DEFAULT_DTYPE = pd.UInt32Dtype()
NONDEFAULT_DTYPES = {
    "anchor.subframe.ok": pd.BooleanDtype(),
    "code": SYNTENY_CATEGORY,
    "fasta_url": pd.StringDtype(),
    "gff_url": pd.StringDtype(),
    "frag.direction": DIRECTIONAL_CATEGORY,
    "frag.id": pd.CategoricalDtype(),
    "frag.is_chr": YES_NO,
    "frag.is_plas": YES_NO,
    "frag.is_scaf": YES_NO,
    "frag.len": pd.UInt64Dtype(),
    "frag.orig_id": pd.StringDtype(),
    "frag.start": pd.UInt64Dtype(),
    "gff.feature": pd.CategoricalDtype(),
    "gff.id": pd.CategoricalDtype(),
    "path": pd.CategoricalDtype(),
    "phy.*": pd.CategoricalDtype(),
    "preference": pd.StringDtype(),
    "prot.m_start": pd.BooleanDtype(),
    "prot.no_stop": pd.BooleanDtype(),
    "prot.seq": pd.StringDtype(),
    "syn.anchor.direction": DIRECTIONAL_CATEGORY,
    "syn.code": SYNTENY_CATEGORY,
    "syn.orthogenomic_pct": "float64",
    "syn.hash.footprint": pd.UInt32Dtype(),
    "syn.hash.direction": DIRECTIONAL_CATEGORY,
    "val": "float64",
    # patterns are matched in order after checking for exact matches
    "patterns": [
        {"start": "phy.", "type": pd.CategoricalDtype()},
        {"start": "pct_", "end": "_pct", "type": "float64"},
        {"start": "memb", "type": pd.StringDtype()},
    ],
}

MEGABYTES = 1024.0 * 1024.0

INSTALL_ENVIRON_VAR = (  # installs go into "/bin" and other subdirs of this directory
    NAME.upper() + "_INSTALL_DIR"
)
if INSTALL_ENVIRON_VAR in os.environ:
    INSTALL_PATH = Path(os.environ[INSTALL_ENVIRON_VAR])
else:
    INSTALL_PATH = Path(sys.executable).parent.parent

# logger class for use in testing


class PrintLogger:
    """This logger only prints, for testing only."""

    def __init__(self, level):
        try:
            self.level = int(level)
        except ValueError:
            if level.lower() == "debug":
                self.level = 10
            elif level.lower() == "info":
                self.level = 20
            elif level.lower() == "warning":
                self.level = 30
            elif level.lower() == "error":
                self.level = 40
            else:
                self.level = 20

    def debug(self, message):
        if self.level <= 10:
            print(f"Debug: {message}")

    def info(self, message):
        if self.level <= 20:
            print(message)

    def warning(self, message):
        if self.level <= 30:
            print(f"Warning: {message}")

    def error(self, message):
        if self.level <= 40:
            print(f"ERROR: {message}")


def is_writable(dev):
    """Returns whether a device is writable or not."""
    try:
        testdir = tempfile.mkdtemp(prefix=NAME + "-", dir=append_slash(dev))
        Path(testdir).rmdir()
    except OSError:
        return False
    return True


# global variables that are set depending on envvars
# Don't use loguru, just print.  Useful for testing.
if "LOG_TO_PRINT" in os.environ:
    logger = PrintLogger(os.environ["LOG_TO_PRINT"])
else:
    logger = loguru_logger
# Update period on spinners.  Also useful for testing.
if "SPINNER_UPDATE_PERIOD" in os.environ:
    try:
        SPINNER_UPDATE_PERIOD = float(os.environ["SPINNER_UPDATE_PERIOD"])
    except ValueError:
        SPINNER_UPDATE_PERIOD = 5.0
else:
    SPINNER_UPDATE_PERIOD = 1.0
# Fast scratch disk (e.g., SSD or /dev/shm), if other than /tmp
if "SCRATCH_DEV" in os.environ and is_writable(os.environ["SCRATCH_DEV"]):
    SCRATCH_DEV = os.environ["SCRATCH_DEV"]
else:
    SCRATCH_DEV = "/tmp"
# Build disk for installer, needs to allow exe bit set
if "BUILD_DEV" in os.environ and is_writable(os.environ["BUILD_DEV"]):
    BUILD_DEV = os.environ["BUILD_DEV"]
elif sys.platform == "linux":
    try:
        BUILD_DEV = MemoryTempfile(
            preferred_paths=["/run/user/{uid}"], fallback=True
        ).get_usable_mem_tempdir_paths()[0]
    except AttributeError:
        BUILD_DEV = "/tmp"
else:
    BUILD_DEV = "/tmp"


def enforce_canonical_dtypes(frame):
    """Enforce that dtypes of columns meet expectations."""
    for col in frame.columns:
        if col.startswith("tmp."):
            continue
        column_type = frame[col].dtype
        should_be_type = DEFAULT_DTYPE
        if col in NONDEFAULT_DTYPES:
            should_be_type = NONDEFAULT_DTYPES[col]
        else:
            for pattern_dict in NONDEFAULT_DTYPES["patterns"]:
                if "start" in pattern_dict:
                    if col.startswith(pattern_dict["start"]):
                        should_be_type = pattern_dict["type"]
                        break
                if "end" in pattern_dict:
                    if col.endswith(pattern_dict["end"]):
                        should_be_type = pattern_dict["type"]
                        break
        try:
            is_correct_type = column_type == should_be_type
        except TypeError:
            is_correct_type = False
        if not is_correct_type:
            try:
                frame[col] = frame[col].astype(should_be_type)
            except (ValueError, TypeError) as cast_err:
                logger.warning(f"Cannot cast {col} to {should_be_type}")
                logger.warning(cast_err)
    return frame


def free_mb(dev):
    """"Return the number of free MB on dev."""
    fs_stats = os.statvfs(dev)
    free_space_mb = int(
        np.rint((fs_stats.f_bsize * fs_stats.f_bfree) / MEGABYTES)
    )
    return free_space_mb


def append_slash(dev):
    """Append a final slash, if needed."""
    if not dev.endswith("/"):
        dev += "/"
    return dev


def disk_usage_mb(pathstr):
    """Calculate the size used in MB by a path."""
    path = Path(pathstr)
    if not path.exists():
        logger.error(f"Path '{path}' does not exist for disk usage")
        return 0
    total_size = np.array(
        [p.stat().st_size for p in path.rglob("*") if not p.is_symlink()]
    ).sum()
    return int(np.rint(total_size / MEGABYTES))


class MinSpaceTracker:
    """Keep track of the minimum space available on a (memory) device."""

    def __init__(self, device):
        """Remember the device and initialize minimum space."""
        self.device = Path(device)
        self.initial_space = free_mb(self.device)
        self.min_space = self.initial_space

    def check(self):
        """Update the minimimum space available."""
        self.min_space = min(self.min_space, free_mb(self.device))

    def report_min(self):
        """Report the minimum space available."""
        return self.min_space

    def report_used(self):
        """Report change from initial space."""
        return self.initial_space - self.min_space


def cluster_set_name(stem, identity):
    """Get a setname that specifies the %identity value.."""
    if identity == 1.0:
        digits = "10000"
    else:
        digits = f"{identity:.4f}"[2:]
    return f"{stem}-nr-{digits}"


def get_paths_from_file(filepath, must_exist=True):
    """Given a string filepath,, return the resolved path and parent."""
    inpath = Path(filepath).expanduser().resolve()
    if must_exist and not inpath.exists():
        raise FileNotFoundError(filepath)
    dirpath = inpath.parent
    return inpath, dirpath


class TrimmableMemoryMap:
    """A memory-mapped file that can be resized at the end."""

    def __init__(self, filepath, access=mmap.ACCESS_WRITE):
        """Open the memory-mapped file."""
        self.orig_size = None
        self.size = None
        self.map_obj = None
        self.access = access
        self.filehandle = open(filepath, "r+b")

    def trim(self, start, end):
        """Trim the memory map and mark the nex size."""
        self.map_obj.move(start, end, self.orig_size - end)
        self.size -= end - start
        return self.size

    @contextlib.contextmanager
    def map(self):
        """Open a memory-mapped view of filepath."""
        try:
            self.map_obj = mmap.mmap(
                self.filehandle.fileno(), 0, access=self.access
            )
            self.orig_size = self.map_obj.size()
            self.size = self.orig_size
            yield self.map_obj
        finally:
            if self.access == mmap.ACCESS_WRITE:
                self.map_obj.flush()
                self.map_obj.close()
                self.filehandle.truncate(self.size)
                self.filehandle.close()
            else:
                self.map_obj.close()
                self.filehandle.close()


def dotpath_to_path(dotpath):
    """Return a dot-separated pathstring as a path."""
    return Path("/".join(dotpath.split(".")))


def fasta_records(filepath):
    """Count the number of records in a FASTA file."""
    count = 0
    next_pos = 0
    angle_bracket = bytes(">", "utf-8")
    memory_map = TrimmableMemoryMap(filepath, access=mmap.ACCESS_READ)
    with memory_map.map() as mem_map:
        size = memory_map.size
        next_pos = mem_map.find(angle_bracket, next_pos)
        while next_pos != -1 and next_pos < size:
            count += 1
            next_pos = mem_map.find(angle_bracket, next_pos + 1)
    return count, size


def protein_file_stats_filename(setname):
    """Return the name of the protein stat file."""
    if setname is None:
        return "protein_files.tsv"
    return f"{setname}-protein_files.tsv"


def protein_properties_filename(filestem):
    """Return the name of the protein properties file."""
    if filestem is None:
        return "proteins.tsv"
    return f"{filestem}-proteins.tsv"


def homo_degree_dist_filename(filestem):
    """Return the name of the homology degree distribution file."""
    return f"{filestem}-degreedist.tsv"


def group_key_filename(members):
    """Return the name of the group key file."""
    return f"groupkeys-{members}.tsv"


def sort_proteome_frame(frame):
    """Sort a proteome frame by preference and frag.max and renumber."""
    frame = frame.copy()
    if frame.index.name == "path":
        frame["path"] = frame.index
    frame.sort_values(
        by=["preference", "frag.max"], ascending=[True, False], inplace=True
    )
    frame["order"] = range(len(frame))
    frame.set_index("order", inplace=True)
    return frame


def remove_tmp_columns(frame):
    """Remove any columns in a data frame that begin with 'tmp.'."""
    drop_cols = [col for col in frame.columns if col.startswith("tmp.")]
    if len(drop_cols) != 0:
        return frame.drop(drop_cols, axis=1)
    return frame


def write_tsv_or_parquet(
    frame,
    filepath,
    compression=DEFAULT_PARQUET_COMPRESSION,
    float_format="%.2f",
    desc=None,
    remove_tmp=True,
    sort_cols=True,
    enforce_types=True,
):
    """Write either a TSV or a parquet file by file extension."""
    filepath = Path(filepath)
    ext = filepath.suffix.lstrip(".")
    if desc is not None:
        file_desc = f"{desc} file"
        logger.debug(f'Writing {file_desc} "{filepath}')
    if remove_tmp:
        frame = remove_tmp_columns(frame)
    if enforce_types:
        frame = enforce_canonical_dtypes(frame)
    if sort_cols:
        frame = frame[sorted(frame.columns)]
    if ext in PARQUET_EXTENSIONS:
        frame.to_parquet(filepath, compression=compression)
    elif ext in TSV_EXTENSIONS:
        frame.to_csv(filepath, sep="\t", float_format=float_format)
    else:
        logger.error(f"Unrecognized file extension {ext} in {filepath}")
        sys.exit(1)


def read_tsv_or_parquet(filepath):
    """Read either a TSV or a parquet file by file extension."""
    filepath = Path(filepath)
    if not filepath.exists():
        logger.error(f'File "{filepath}" does not exist.')
        sys.exit(1)
    ext = filepath.suffix.lstrip(".")
    if ext in PARQUET_EXTENSIONS:
        return pd.read_parquet(filepath)
    if ext in TSV_EXTENSIONS:
        frame = pd.read_csv(filepath, sep="\t", index_col=0).convert_dtypes()
        return enforce_canonical_dtypes(frame)
    logger.error(f"Unrecognized file extensions {ext} in {filepath}")
    sys.exit(1)


def log_and_add_to_stats(stats, new_stats):
    """Print stats info and write to stats file."""
    with pd.option_context(
        "display.max_rows",
        None,
        "display.max_columns",
        None,
        "display.float_format",
        "{:,.1f}%".format,
    ):
        logger.info(new_stats)
    overlap_cols = list(set(stats.columns) & set(new_stats.columns))
    return pd.concat([stats.drop(columns=overlap_cols), new_stats], axis=1)


def bool_to_y_or_n(bool_arr):
    """Convert boolean to T/F value"""
    boolean_dict = {True: "y", False: "n"}
    bool_ser = pd.Series(bool_arr)
    return bool_ser.map(boolean_dict)


def y_or_n_to_bool(bool_ser):
    """Convert boolean to T/F value"""
    tf_dict = {"y": True, False: "n"}
    return bool_ser.map(tf_dict).astype(bool)


def hash_array(kmer):
    """Return a hash of a numpy array."""
    return xxhash.xxh32_intdigest(kmer.tobytes())


def calculate_adjacency_group(index_series, frag_series):
    """Calculate an adjacency group numger."""
    index_fr = pd.DataFrame({"index": index_series, "fragment": frag_series})
    n_prot = len(index_fr)
    adj_gr_count = 0
    was_adj = False
    index_fr["i"] = range(n_prot)
    adj_group = np.array([np.nan] * n_prot)
    for unused_group, subframe in index_fr.groupby(by=["fragment"]):
        if len(subframe) == 1:
            continue
        last_pos = -2
        last_row = None
        if was_adj:
            adj_gr_count += 1
        was_adj = False
        for unused_i, row in subframe.iterrows():
            row_no = row["i"]
            if row["index"] == last_pos + 1:
                if not was_adj:
                    adj_group[last_row] = adj_gr_count
                was_adj = True
                adj_group[row_no] = adj_gr_count
            else:
                if was_adj:
                    adj_gr_count += 1
                    was_adj = False
            last_pos = row["index"]
            last_row = row_no
    if was_adj:
        adj_gr_count += 1
    adj_arr = pd.Series(
        adj_group, dtype=pd.UInt32Dtype(), index=index_series.index
    )
    n_adj = n_prot - adj_arr.isnull().sum()
    return n_adj, adj_gr_count, adj_arr


def get_bin_paths():
    "Return paths from envvar PATH with binary prepended."
    search_paths = []
    for pathstr in os.environ.get("PATH", "").split(os.pathsep):
        test_path = Path(pathstr).expanduser()
        if test_path.is_dir():
            search_paths.append(str(test_path))
    if len(search_paths) == 0:
        logger.warning("Empty PATH environmental variable")
    venv_bin_path = str(INSTALL_PATH.joinpath("bin"))
    if venv_bin_path not in search_paths:
        search_paths = [venv_bin_path] + search_paths
    return tuple(search_paths)


SEARCH_PATHS = get_bin_paths()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['azulejo', 'azulejo.installer_data.dagchainer-tool']

package_data = \
{'': ['*'],
 'azulejo': ['examples/glycine7/*',
             'installer_data/blast-longids/*',
             'installer_data/blast-longids/linux/*',
             'installer_data/blast-longids/macos/*',
             'installer_data/mmseqs/linux/*',
             'installer_data/mmseqs/macos/*',
             'installer_data/muscle/*',
             'installer_data/muscle/linux/*',
             'installer_data/muscle/macos/*',
             'installer_data/usearch/*',
             'installer_data/usearch/linux/*',
             'installer_data/usearch/macos/*']}

install_requires = \
['amas>=1.0,<2.0',
 'attrs',
 'beautifulsoup4>=4.9.3,<5.0.0',
 'biopython>=1.78,<2.0',
 'click>=7.1.2,<8.0.0',
 'click_loguru>=1.3.0,<2.0.0',
 'click_plugins>=1.1.1,<2.0.0',
 'dask[bag]>=2.15.0,<3.0.0',
 'fastaq>=0.30,<0.31',
 'gffpandas-ncgr>=1.2.1,<2.0.0',
 'matplotlib>=3.2.1,<4.0.0',
 'memory-tempfile>=2.2.3,<3.0.0',
 'networkx>=2.4,<3.0',
 'numpy>=1.18.3,<2.0.0',
 'packaging>=20.7,<21.0',
 'pandas>=1.1.0,<2.0.0',
 'paramiko>=2.7.1,<3.0.0',
 'pathvalidate>=2.3.0,<3.0.0',
 'phylopandas>=0.8.0,<0.9.0',
 'progressbar2>=3.53.1,<4.0.0',
 'pyarrow>=2.0.0,<3.0.0',
 'requests_download>=0.1.2,<0.2.0',
 'seaborn>=0.11.0,<0.12.0',
 'sh>=1.14.1,<2.0.0',
 'smart-open>=2.0.0,<3.0.0',
 'toml>=0.10.1,<0.11.0',
 'uri>=2.0.1,<3.0.0',
 'xxhash>=2.0.0,<3.0.0']

entry_points = \
{'console_scripts': ['azulejo = azulejo:cli']}

setup_kwargs = {
    'name': 'azulejo',
    'version': '0.10.0',
    'description': 'tile phylogenetic space with subtrees',
    'long_description': '.. epigraph:: azulejo\n              noun INFORMAL\n              a glazed tile, usually blue, found on the inside of churches and palaces in Spain and Portugal.\n\nazulejo\n=======\n``azulejo`` azulejo combines homology and synteny information to\ntile phylogenetic space.\nThe inputs to ``azulejo`` are FASTA files of nucleotide-space\nsequences of primary-transcript protein genes and their associated GFF files.\nOutputs are sets of proxy gene fragments chosen for \ntheir concordance in multiple sequence alignments, along with\nsubtrees.\n\nPrerequisites\n-------------\nPython 3.7 or greater is required. ``azulejo`` is tested under Linux \nusing Python 3.8 and 3.9 and under MacOS Big Sur using XCode command-line tools\nsystem Python (currently 3.8). Mac users should see the `instructions\non configuring their systems <macos.rst>`_.  Installation on BSD is not\nsupported because many of the python dependencies lack BSD wheels.\n\nWe recommend you install ``azulejo`` into its own virtual environment due\nto the large number of python dependencies.  The easiest way for most \nusers to install and maintain up-to-date virtual environments is via the\ntool `pipx<https://pipxproject.github.io/pipx/>`_.  If your system does\nnot have ``pipx`` installed, you can do so via the commands::\n\n        python3 -m pip install --user --upgrade pip\n        python3 -m pip install --user --upgrade pipx\n        python3 -m pipx ensurepath\n\nFollow any instructions that the last command produces about starting a new\nshell if necessary.  \n\nIf you choose to have ``azulejo`` compile and install its binary dependencies,\nyou will need compilers, ``make``, and ``cmake``  and standard headers\nfor ``zlib`` and ``bz2``.  All linux systems configured for development will have\nthese available.  We test compilation under gcc version 10.2 on linux and\nclang 12.0.0 on MacOS.  We use program-guided optimization for one of the\nbinary dependencies, and we believe that gcc 10 does a much better job\nof optimization than gcc 9, so it may benefit you to upgrade your compiler\nif needed.\n\nInstallation for Users\n----------------------\nOnce the prerequisite has been met, you may then install ``azulejo`` \nin its own virtual environment by issuing the command::\n\n        pipx install azulejo\n\n``azulejo`` contains some long commands and many options.  To enable command-line\ncompletion for ``azulejo`` commands, execute the following command if you are using\n``bash`` as your shell: ::\n\n    eval "$(_AZULEJO_COMPLETE=source_bash azulejo)"\n\nThen you should run ``azulejo install`` and check the versions of all binary\ndependencies that may installed system-wide.\n\n\nEnvironmental Variables\n-----------------------\n``azulejo`` recognizes the following environmental variables:\n\n* AZULEJO_INSTALL_DIR\nThis is a writable directory for installation of binary dependencies.  Binaries\nwill go into the ``bin`` directory.  The default is the virtual environment\ndirectory.\n\n* BUILD_DEV\nThis is the directory used for building binary dependencies.  Default is the\nfirst memory device found for linux (e.g., `/run/shm`) or `/tmp` for MacOS.\nSet this if compilation fails because it runs out of memory.\n\n* SCRATCH_DEV\nThis is the directory used for temporary merging of lists.  The default is\n`/tmp`, but you may set it to a fast memory based device if you have enough\nmemory.\n\n* MAKEOPTS\nThese are the arguments to the ``make`` and ``make install`` commands when\nbuilding dependencies.  It\'s good to set this to the number of processors\non your system via the command ``export MAKEOPTS="-j $(nproc)"`` to speed\nup installation.  The only time this variable is used is during\n``azulejo install``.\n\n* SPINNER_UPDATE_PERIOD\nThis is the number of seconds between updates of the spinner.  This\ndefaults to 1, but it is advisable to set it higher for automated testing\nso as not to exceed logfile character limits.\n\n* LOG_TO_PRINT\nIf set to a log level such as ``info``, the logger will be a simple print without using the more\ncomplex functions of ``loguru`` such as colors and logging to files.\nThis is sometimes useful in automated testing.\n\n\n\nInstallation of Binary Dependencies\n-----------------------------------\n``azulejo`` requires `MMseqs <https://github.com/soedinglab/MMseqs2>`_ \nfor homology clustering and `MUSCLE <https://www.drive5.com/muscle/downloads.htm>`_\nfor sequence alignment and initial tree-building.\n``azulejo`` installs binaries into the virtualenv by default, so\nany systemwide installations of these packages will not get clobbered by the install.\nIn particular, ``muscle`` is PGO-optimized, which gives nearly a factor of 2 higher\nperformance than prebuilt binaries.  We recommand you set ``MAKEOPTS`` as explained\nabove, then issue the command ``azulejo install all`` to ensure you get correct versions\noptimized for your hardware.\n\n\nThere are three optional dependencies that can be installed via ``azulejo install`` \nthat are of interest only to a small subset of users who wish to compare against\nother homology clustering and synteny methods.  \n`usearch <https://www.drive5.com/usearch/download.html>`_ \nis a licensed homology clustering program that is free for individual, non-commercial\nuse that can be downloaded and installed by the ``azulejo install usearch``\ncommand after accepting the license terms.  ``azulejo install dagchainer-tool`` gets you\na somewhat crude Bash script that uses BLAST homology clustering followed by \nsynteny calculation via `DAGchainer <https://dagchainer.sourceforge.net>`_.  \n``dagchainer-tool`` will need the dependency of ``perl`` with ``bioperl`` installed.\n``dagchainer_tool`` increases the sequence ID length as part of its processing, so\nif any of your sequence IDS are longer than about 30 characters, they will violate BLAST\'s\nhard limit of 50 characters in sequence ID fields.  In that case you will need\nto install a patched version of BLAST using the command ``azulejo install blast-longids``.\n\nInstallation For Developers\n---------------------------\nIf you plan to develop ``azulejo``, you\'ll need to install\nthe `poetry <https://python-poetry.org>`_ dependency manager.\nIf you haven\'t previously installed ``poetry``, execute the command: ::\n\n    curl -sSL https://raw.githubusercontent.com/python-poetry/poetry/master/get-poetry.py | python\n\nNext, get the master branch from GitHub ::\n\n\tgit clone https://github.com/legumeinfo/azulejo.git\n\nChange to the ``azulejo/`` directory and install with poetry: ::\n\n\tpoetry install -v\n\nRun ``azulejo`` with ``poetry``: ::\n\n    poetry run azulejo\n\nUsage\n-----\nInstallation puts a single script called ``azulejo`` in your path.  The usage format is::\n\n    azulejo [GLOBALOPTIONS] COMMAND [COMMANDOPTIONS][ARGS]\n\n\nMaster Input File\n-----------------\n``azulejo`` uses a configuration file in `TOML  <https://github.com/toml-lang/toml>`_\nformat as the master input that associates files with phylogeny.  The format of this file\nis the familiar headings in square brackets followed by configuration values::\n\n    [glycines]\n    rank = "genus"\n    name = "Glycine"\n\n    [glycines.glyso]\n    rank = "species"\n    name = "Glycine soja"\n\n    [glycines.glyso.PI483463]\n    rank = "strain"\n    gff = "glyso.PI483463.gnm1.ann1.3Q3Q.gene_models_main.gff3.gz"\n    fasta = "glyso.PI483463.gnm1.ann1.3Q3Q.protein_primaryTranscript.faa.gz"\n    uri = "https://v1.legumefederation.org/data/index/public/Glycine_soja/PI483463.gnm1.ann1.3Q3Q/"\n    comments = """\n    Glycine soja accession PI 483463 has been identified as being unusually\n    salt-tolerant (Lee et al., 2009)."""\n\n\n* [headings]\n    There can be only one top-level heading, and that will be the name of the\n    resulting output set.  This name will be the name of an output directory that will be\n    created in the current working directory, so this heading (and all subheadings) must\n    obey UNIX filesystem naming rules or an error will result.  Each heading level\n    (indicated by a ".") will result in another taxonomic level and another directory level\n    in the output directory.  Depths do not need to be consistent.\n\n* rank\n    Each level defined must have a ``rank`` defined, and that rank must match one of the\n    taxonomic ranks defined by ``azulejo``, which you can view and test using the\n    ``check-taxonomic-rank`` command.   There are 24 major taxonomic ranks, each of which\n    may be modified by 16 different prefixes for a total of 174 taxonomic levels (some of\n    which are synonoymous).\n\n* name\n    Each level may (and usually should) have a ``name`` defined.  This name is intended\n    to be human-readable with no restrictions on the characters used, but it goes into\n    plot legends in places, so it\'s best to not make it too long. If the name is not specified,\n    it will be taken from the level name enclosed in single quotes (e.g., \'PI483463\' for the\n    example above).\n\n* fasta\n    If the level specifies a genome, it must have a ``fasta`` entry corresponding\n    to the name of the *protein* FASTA file.  In eukaryotes, the FASTA file should be a\n    file of primary (generally longest) protein transcripts, if available, rather than all protein\n    transcripts (i.e., not including splice variants). Sequences will be cleaned of dashes, stops,\n    and other out-of-alphabet characters.  Ambiguous residues at the beginnings and ends of\n    sequences will be trimmed. Zero-length sequences will be discarded, which can result in a\n    smaller number of sequences out.  These files may be compressed, with extensions ``.gz`` or\n    ``.bz2``.\n\n* gff\n    If the level specifies a genome, it must have a ``gff`` entry corresponding\n    to a version 3 Genome Feature File (GFF3) containing ``CDS`` entries with ID values\n    matching those IDs in the FASTA file.  The same compression extensions as for\n    ``fasta`` entries apply.  If the ``SOURCE`` fields in those CDS entries\n    (which contain the names of the DNA fragments such as scaffolds that the CDS came from)\n    contain dot-separated components, those components that are identical across the entire\n    file will be discarded by default.  There is an opportunity later in the process to\n    remap DNA source names to a common dictionary for comparison among chromosomes and\n    plastids.\n\n* uri\n    This optional field may contain a a uniform resource identifier such as\n    ``https://sitename/dir/``.  ``azulejo`` uses `smart-open <https://www.pypi.org/project/smart-open/>`_\n    for doing transparent on-the-fly decompression from a variety of file systems\n    including HTTPS, HDFS, SSH, and SFTP (but not FTP).\n    If this field is not supplied, local file access is assumed with paths relative to\n    the current working directory. The URI will be prepended to ``fasta``\n    and ``gff`` paths, allowing for convenient downloading on-the-fly from sites such as\n    LegumeInfo or GenBank.   Downloads are not cached, so if you intend to run ``azulejo``\n    multiple times on the same input data, you will save time by downloading and uncompressing\n    files to local storage.\n\n* preference\n    This optional field may be used to override the genome preference heuristic\n    that is the fall-thru preference after proxy-gene heuristics have been applied.  This is an integer\n    value, with lower integers getting the highest priority.  Set this value to zero if you\n    know in advance that one of the input genomes is considered the reference genome and,\n    all things being equal, you would prefer to select proxy genes from this genome.  You\n    may also set these preference values later, after the default genome preference (genomes\n    will be preferred in order of the most genes in a single DNA fragment) has already been\n    applied, but before proxy gene selection.\n\n* other info\n    A design goal for ``azulejo`` was to not lose metadata, even if it\n    was not used by ``azulejo`` itself, while keeping metadata out of file names.\n    As an aid in that goal, for each (sub)heading level/output directory, ``azulejo``\n    creates a JSON file named ``node_properties.json`` at each node in the output\n    hierarchy that containing all information from this file as well as other information\n    calculated at ingestion time by ``azulejo``.  You may specify any additional data you would\n    like to pass along (e.g., for later use in a web page) and it will be translated from TOML\n    to JSON and passed along, such as the multi-line ``comments`` field in the example.\n    Examples of useful metadata that may be easier to enter at ingestion time than to\n    garner later include taxon IDs of the level and its parent, common names, URLs of\n    papers describing the genome, and geographic origin of the sample.\n\nA copy of the input file will be saved in the output directory under the name ``input.toml``.\nSee the examples in the ``tests/testdata`` repository directory for examples of input data.\n\nGlobal Options\n--------------\nThe following options are global in scope and, if used must be placed before\n``COMMAND``:\n\n============================= ===========================================\n   -v, --verbose              Log debugging info to stderr.\n   -q, --quiet                Suppress logging to stderr.\n   --no-logfile               Suppress logging to file.\n   -e, --warnings_as_errors   Treat warnings as fatal (for testing).\n============================= ===========================================\n\nCommands\n--------\nA listing of commands is available via ``azulejo --help``.\nThe currently implemented commands are, in the order they will normally be run:\n\n========================= ==================================================\n  install                 Check for/install binary dependencies.\n  ingest                  Marshal protein and genome sequence information.\n  homology                Calculate homology clusters, MSAs, trees.\n  synteny                 Calculate synteny anchors.\n  proxy-genes             Calculate a set of proxy genes from synteny files.\n  parquet-to-tsv          Reads parquet file, writes tsv.\n========================= ==================================================\n\n``azulejo`` stores most intermediate results in the Parquet format with\nextension ``.parq``.  These binary files are compressed and typically can\nbe read more than 30X faster than the tab-separated-value (TSV) files they\ncan be interconverted with.  In addition, Parquet files do not lose metadata\nsuch as binary representation sizes.\n\nEach command has its ``COMMANDOPTIONS``, which may be listed with: ::\n\n    azulejo COMMAND --help\n\nProject Status\n--------------\n+-------------------+-------------+------------+\n| Latest Release    | |pypi|      | |azulejo|  |\n+-------------------+-------------+            +\n| Activity          | |repo|      |            |\n+-------------------+-------------+            +\n| Downloads         | |downloads| |            |\n+-------------------+-------------+            +\n| Download Rate     | |dlrate|    |            |\n+-------------------+-------------+            +\n| License           | |license|   |            |\n+-------------------+-------------+            +\n| Code Grade        | |codacy|    |            |\n+-------------------+-------------+            +\n| Coverage          | |coverage|  |            |\n+-------------------+-------------+            +\n| Travis Build      | |travis|    |            |\n+-------------------+-------------+            +\n| Issues            | |issues|    |            |\n+-------------------+-------------+            +\n| Code Style        | |black|     |            |\n+-------------------+-------------+------------+\n\n\n.. |azulejo| image:: docs/azulejo.jpg\n     :target: https://en.wikipedia.org/wiki/Azulejo\n     :alt: azulejo Definition\n\n.. |black| image:: https://img.shields.io/badge/code%20style-black-000000.svg?style=flat-square\n    :target: https://github.com/psf/black\n    :alt: Black is the uncompromising Python code formatter\n\n.. |pypi| image:: https://img.shields.io/pypi/v/azulejo.svg\n    :target: https://pypi.python.org/pypi/azulejo\n    :alt: Python package\n\n.. |repo| image:: https://img.shields.io/github/last-commit/legumeinfo/azulejo\n    :target: https://github.com/legumeinfo/azulejo\n    :alt: GitHub repository\n\n.. |license| image:: https://img.shields.io/badge/License-BSD%203--Clause-blue.svg\n    :target: https://github.com/legumeinfo/azulejo/blob/master/LICENSE\n    :alt: License terms\n\n.. |rtd| image:: https://readthedocs.org/projects/azulejo/badge/?version=latest\n    :target: http://azulejo.readthedocs.io/en/latest/?badge=latest\n    :alt: Documentation Server\n\n.. |travis| image:: https://img.shields.io/travis/legumeinfo/azulejo.svg\n    :target:  https://travis-ci.org/legumeinfo/azulejo\n    :alt: Travis CI\n\n.. |codacy| image:: https://api.codacy.com/project/badge/Grade/99549f0ed4e6409e9f5e80a2c4bd806b\n    :target: https://www.codacy.com/app/joelb123/azulejo?utm_source=github.com&amp;utm_medium=referral&amp;utm_content=legumeinfo/azulejo&amp;utm_campaign=Badge_Grade\n    :alt: Codacy.io grade\n\n.. |coverage| image:: https://codecov.io/gh/legumeinfo/azulejo/branch/master/graph/badge.svg\n    :target: https://codecov.io/gh/legumeinfo/azulejo\n    :alt: Codecov.io test coverage\n\n.. |issues| image:: https://img.shields.io/github/issues/LegumeFederation/lorax.svg\n    :target:  https://github.com/legumeinfo/azulejo/issues\n    :alt: Issues reported\n\n.. |requires| image:: https://requires.io/github/legumeinfo/azulejo/requirements.svg?branch=master\n     :target: https://requires.io/github/legumeinfo/azulejo/requirements/?branch=master\n     :alt: Requirements Status\n\n.. |dlrate| image:: https://img.shields.io/pypi/dm/azulejo\n    :target: https://pypistats.org/packages/azulejo\n    :alt: Download stats\n\n.. |downloads| image:: https://pepy.tech/badge/azulejo\n    :target: https://pepy.tech/project/azulejo\n    :alt: Download stats\n',
    'author': 'Joel Berendzen',
    'author_email': 'joelb@ncgr.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/legumeinfo/azulejo',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-
"""Core logic for computing subtrees."""

# standard library imports
import contextlib
import os
import sys
from collections import Counter
from collections import OrderedDict
from itertools import chain
from itertools import combinations
from pathlib import Path

# third-party imports
import networkx as nx
import numpy as np
import pandas as pd
from Bio import SeqIO

# first-party imports
import sh

# module imports
from .common import CLUSTER_HIST_FILE
from .common import NAME
from .common import SEARCH_PATHS
from .common import cluster_set_name
from .common import fasta_records
from .common import get_paths_from_file
from .common import homo_degree_dist_filename
from .common import logger
from .common import protein_properties_filename
from .common import write_tsv_or_parquet
from .protein import Sanitizer

# global constants
STATFILE_SUFFIX = f"-{NAME}_stats.tsv"
ANYFILE_SUFFIX = f"-{NAME}_ids-any.tsv"
ALLFILE_SUFFIX = f"-{NAME}_ids-all.tsv"
CLUSTFILE_SUFFIX = f"-{NAME}_clusts.tsv"
SEQ_FILE_TYPE = "fasta"
UNITS = {
    "Mb": {"factor": 1, "outunits": "MB"},
    "Gb": {"factor": 1024, "outunits": "MB"},
    "s": {"factor": 1, "outunits": "s"},
    "m": {"factor": 60, "outunits": "s"},
    "h": {"factor": 3600, "outunits": "s"},
}
SEQ_IN_LINE = 6
IDENT_STATS_LINE = 7
FIRST_LOG_LINE = 14
LAST_LOG_LINE = 23
STAT_SUFFIXES = ["size", "mem", "time", "memory"]
RENAME_STATS = {
    "throughput": "throughput_seq_s",
    "time": "CPU_time",
    "max_size": "max_cluster_size",
    "avg_size": "avg_cluster_size",
    "min_size": "min_cluster_size",
    "seqs": "unique_seqs",
    "singletons": "singleton_clusters",
}
ID_SEPARATOR = "."
IDENT_LOG_MIN = -3
IDENT_LOG_MAX = 0
FASTA_EXT_LIST = [".faa", ".fa", ".fasta"]
FAA_EXT = "faa"


# helper functions
def read_synonyms(filepath):
    """Read a file of synonymous IDs into a dictionary."""
    synonym_dict = {}
    try:
        synonym_frame = pd.read_csv(filepath, sep="\t")
    except FileNotFoundError:
        logger.error(f'Synonym tsv file "{filepath}" does not exist')
        sys.exit(1)
    except pd.errors.EmptyDataError:
        logger.error(f'Synonym tsv "{filepath}" is empty')
        sys.exit(1)
    if len(synonym_frame) > 0:
        if "#file" in synonym_frame:
            synonym_frame.drop("#file", axis=1, inplace=True)
        key = list({("Substr", "Dups")}.intersection({synonym_frame.columns}))[
            0
        ]
        for group in synonym_frame.groupby("id"):
            synonym_dict[group[0]] = group[1][key]
    return synonym_dict


def parse_usearch_log(filepath, rundict):
    """Parse the usearch log file into a stats dictionary."""
    with filepath.open() as logfile:
        for lineno, line in enumerate(logfile):
            if lineno < FIRST_LOG_LINE:
                if lineno == SEQ_IN_LINE:
                    split = line.split()
                    rundict["seqs_in"] = int(split[0])
                    rundict["singleton_seqs_in"] = int(split[4])
                if lineno == IDENT_STATS_LINE:
                    split = line.split()
                    rundict["max_identical_seqs"] = int(split[6].rstrip(","))
                    rundict["avg_identical_seqs"] = float(split[8])
                continue
            if lineno > LAST_LOG_LINE:
                break
            split = line.split()
            if split:
                stat = split[0].lower()
                if split[1] in STAT_SUFFIXES:
                    stat += "_" + split[1]
                    val = split[2]
                else:
                    val = split[1].rstrip(",")
                # rename poorly-named stats
                stat = RENAME_STATS.get(stat, stat)
                # strip stats with units at the end
                conversion_factor = 1
                for unit in UNITS:
                    if val.endswith(unit):
                        val = val.rstrip(unit)
                        conversion_factor = UNITS[unit]["factor"]
                        stat += "_" + UNITS[unit]["outunits"]
                        break
                # convert string values to int or float where possible
                try:
                    val = int(val)
                    val *= conversion_factor
                except ValueError:
                    try:
                        val = float(val)
                        val *= conversion_factor
                    except ValueError:
                        pass
                rundict[stat] = val


@contextlib.contextmanager
def in_working_directory(path):
    """Change working directory and return to previous wd on exit."""
    original_cwd = Path.cwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(original_cwd)


def get_fasta_ids(fasta):
    """Get the IDS from a FASTA file."""
    idset = set()
    with fasta.open() as fasta_fh:
        for line in fasta_fh:
            if line.startswith(">"):
                idset.add(line.split()[0][1:])
    return list(idset)


def parse_chromosome(ident):
    """Parse chromosome identifiers."""
    # If ident contains an underscore, work on the
    # last part only (e.g., MtrunA17_Chr4g0009691)
    undersplit = ident.split("_")
    if len(undersplit) > 1:
        ident = undersplit[-1].upper()
        if ident.startswith("CHR"):
            ident = ident[3:]
    # Chromosome numbers are integers suffixed by 'G'
    try:
        chromosome = "Chr" + str(int(ident[: ident.index("G")]))
    except ValueError:
        chromosome = None
    return chromosome


def parse_subids(ident):
    """Parse the subidentifiers from identifiers."""
    subids = ident.split(ID_SEPARATOR)
    subids += [
        chromosome
        for chromosome in [parse_chromosome(ident) for ident in subids]
        if chromosome is not None
    ]
    return subids


def parse_clusters(outdir, delete=True, count_clusters=True, synonyms=None):
    """Parse clusters, counting occurrances."""
    if synonyms is None:
        synonyms = {}
    cluster_list = []
    id_list = []
    degree_list = []
    size_list = []
    degree_counter = Counter()
    any_counter = Counter()
    all_counter = Counter()
    graph = nx.Graph()
    for fasta in outdir.glob("*"):
        cluster_id = int(fasta.name)
        ids = get_fasta_ids(fasta)
        if len(synonyms) > 0:
            syn_ids = set(ids).intersection(synonyms.keys())
            for i in syn_ids:
                ids.extend(synonyms[i])
        n_ids = len(ids)
        degree_list.append(n_ids)
        degree_counter.update({n_ids: 1})
        id_list += ids
        cluster_list += [cluster_id] * n_ids
        size_list += [n_ids] * n_ids
        # Do 'any' and 'all' counters
        id_counter = Counter()
        id_counter.update(
            chain.from_iterable(
                [parse_subids(cluster_id) for cluster_id in ids]
            )
        )
        if count_clusters:
            any_counter.update(id_counter.keys())
            all_counter.update(
                [
                    cluster_id
                    for cluster_id in id_counter.keys()
                    if id_counter[cluster_id] == n_ids
                ]
            )
        elif n_ids > 1:
            any_counter.update({s: n_ids for s in id_counter.keys()})
            all_counter.update(
                {
                    cluster_id: n_ids
                    for cluster_id in id_counter.keys()
                    if id_counter[cluster_id] == n_ids
                }
            )
        # Do graph components
        graph.add_nodes_from(ids)
        if n_ids > 1:
            edges = combinations(ids, 2)
            graph.add_edges_from(edges, weight=n_ids)
        if delete:
            fasta.unlink()
    if delete:
        outdir.rmdir()
    return (
        graph,
        cluster_list,
        id_list,
        size_list,
        degree_list,
        degree_counter,
        any_counter,
        all_counter,
    )


def prettyprint_float(val, digits):
    """Print a floating-point value in a nice way."""
    format_string = "%." + f"{digits:d}" + "f"
    return (format_string % val).rstrip("0").rstrip(".")


def homology_cluster(
    seqfile,
    identity,
    delete=True,
    write_ids=False,
    do_calc=True,
    min_id_freq=0,
    substrs=None,
    dups=None,
    cluster_stats=True,
    outname=None,
    click_loguru=None,
):
    """Cluster at a global sequence identity threshold."""
    try:
        usearch = sh.Command("usearch", search_paths=SEARCH_PATHS)
    except sh.CommandNotFound:
        logger.error("usearch must be installed first.")
        sys.exit(1)
    try:
        inpath, dirpath = get_paths_from_file(seqfile)
    except FileNotFoundError:
        logger.error(f'Input file "{seqfile}" does not exist!')
        sys.exit(1)
    stem = inpath.stem
    dirpath = inpath.parent
    if outname is None:
        outname = cluster_set_name(stem, identity)
    outdir = f"{outname}/"
    logfile = f"{outname}.log"
    outfilepath = dirpath / outdir
    logfilepath = dirpath / logfile
    histfilepath = dirpath / homo_degree_dist_filename(outname)
    gmlfilepath = dirpath / f"{outname}.gml"
    statfilepath = dirpath / f"{outname}-stats.tsv"
    anyfilepath = dirpath / f"{outname}-anyhist.tsv"
    allfilepath = dirpath / f"{outname}-allhist.tsv"
    idpath = dirpath / f"{outname}-ids.tsv"
    if identity == 0.0:
        identity_string = "Minimum"
    else:
        identity_string = f"{prettyprint_float(identity *100, 2)}%"
    logger.info(f'{identity_string} sequence identity cluster "{outname}":')
    if not delete:
        logger.debug(f"Cluster files will be kept in {logfile} and {outdir}")
    if cluster_stats and write_ids:
        logger.debug(
            f"File of cluster ID usage will be written to {anyfilepath} and"
            f" {allfilepath}"
        )
    if not do_calc:
        if not logfilepath.exists():
            logger.error("Previous results must exist, rerun with --do_calc")
            sys.exit(1)
        logger.debug("Using previous results for calculation")
    if min_id_freq:
        logger.debug(
            "Minimum number of times ID's must occur to be counted:"
            f" {min_id_freq}"
        )
    synonyms = {}
    if substrs is not None:
        logger.debug(f"using duplicates in {dirpath / dups}")
        synonyms.update(read_synonyms(dirpath / substrs))
    if dups is not None:
        logger.debug(f"using duplicates in {dirpath/dups}")
        synonyms.update(read_synonyms(dirpath / dups))
    click_loguru.elapsed_time("Clustering")
    if do_calc:
        #
        # Delete previous results, if any.
        #
        if outfilepath.exists() and outfilepath.is_file():
            outfilepath.unlink()
        elif outfilepath.exists() and outfilepath.is_dir():
            for file in outfilepath.glob("*"):
                file.unlink()
        else:
            outfilepath.mkdir()
        #
        # Do the calculation.
        #
        with in_working_directory(dirpath):
            output = usearch(
                [
                    "-cluster_fast",
                    seqfile,
                    "-id",
                    identity,
                    "-clusters",
                    outdir,
                    "-log",
                    logfile,
                ]
            )
            logger.debug(output)
    run_stat_dict = OrderedDict([("divergence", 1.0 - identity)])
    parse_usearch_log(logfilepath, run_stat_dict)
    run_stats = pd.DataFrame(
        list(run_stat_dict.items()), columns=["stat", "val"]
    )
    run_stats.set_index("stat", inplace=True)
    write_tsv_or_parquet(run_stats, statfilepath)
    if delete:
        logfilepath.unlink()
    if not cluster_stats:
        file_sizes = []
        file_names = []
        record_counts = []
        logger.debug("Ordering clusters by number of records and size.")
        for fasta_path in outfilepath.glob("*"):
            records, size = fasta_records(fasta_path)
            if records == 1:
                fasta_path.unlink()
                continue
            file_names.append(fasta_path.name)
            file_sizes.append(size)
            record_counts.append(records)
        file_frame = pd.DataFrame(
            list(zip(file_names, file_sizes, record_counts)),
            columns=["name", "size", "seqs"],
        )
        file_frame.sort_values(
            by=["seqs", "size"], ascending=False, inplace=True
        )
        file_frame["idx"] = range(len(file_frame))
        for unused_id, row in file_frame.iterrows():
            (outfilepath / row["name"]).rename(
                outfilepath / f'{row["idx"]}.fa'
            )
        file_frame.drop(["name"], axis=1, inplace=True)
        file_frame.set_index("idx", inplace=True)
        # write_tsv_or_parquet(file_frame, "clusters.tsv")
        # cluster histogram
        cluster_hist = pd.DataFrame(file_frame["seqs"].value_counts())
        cluster_hist.rename(columns={"seqs": "clusters"}, inplace=True)
        cluster_hist.index.name = "n"
        cluster_hist.sort_index(inplace=True)
        total_seqs = sum(file_frame["seqs"])
        n_clusters = len(file_frame)
        cluster_hist["pct_clusts"] = (
            cluster_hist["clusters"] * 100.0 / n_clusters
        )
        cluster_hist["pct_seqs"] = (
            cluster_hist["clusters"] * cluster_hist.index * 100.0 / total_seqs
        )
        cluster_hist.to_csv(CLUSTER_HIST_FILE, sep="\t", float_format="%06.3f")
        return n_clusters, run_stats, cluster_hist
    (
        cluster_graph,
        clusters,
        ids,
        sizes,
        unused_degrees,
        degree_counts,
        any_counts,
        all_counts,
    ) = parse_clusters(  # pylint: disable=unused-variable
        outfilepath, delete=delete, synonyms=synonyms
    )
    #
    # Write out list of clusters and ids.
    #
    id_frame = pd.DataFrame.from_dict(
        {
            "id": ids,
            "hom.cluster": pd.array(clusters, dtype=pd.UInt32Dtype()),
            "siz": sizes,
        }
    )
    id_frame.sort_values("siz", ascending=False, inplace=True)
    id_frame = id_frame.reindex(
        ["hom.cluster", "siz", "id"],
        axis=1,
    )
    id_frame.reset_index(inplace=True)
    id_frame.drop(["index"], axis=1, inplace=True)
    id_frame.to_csv(idpath, sep="\t")
    del ids, clusters, sizes, id_frame
    click_loguru.elapsed_time("graph")
    #
    # Write out degree distribution.
    #
    cluster_hist = pd.DataFrame(
        list(degree_counts.items()), columns=["degree", "clusters"]
    )
    cluster_hist.sort_values(["degree"], inplace=True)
    cluster_hist.set_index("degree", inplace=True)
    total_clusters = cluster_hist["clusters"].sum()
    cluster_hist["pct_total"] = (
        cluster_hist["clusters"] * 100.0 / total_clusters
    )
    cluster_hist.to_csv(histfilepath, sep="\t", float_format="%06.3f")
    del degree_counts
    #
    # Do histograms of "any" and "all" id usage in cluster
    #
    hist_value = f"{identity:f}"
    any_hist = pd.DataFrame(
        list(any_counts.items()), columns=["id", hist_value]
    )
    any_hist.set_index("id", inplace=True)
    any_hist.sort_values(hist_value, inplace=True, ascending=False)
    all_hist = pd.DataFrame(
        list(all_counts.items()), columns=["id", hist_value]
    )
    all_hist.set_index("id", inplace=True)
    all_hist.sort_values(hist_value, inplace=True, ascending=False)
    if min_id_freq:
        any_hist = any_hist[any_hist[hist_value] > min_id_freq]
        all_hist = all_hist[all_hist[hist_value] > min_id_freq]
    if write_ids:
        any_hist.to_csv(anyfilepath, sep="\t")
        all_hist.to_csv(allfilepath, sep="\t")
    #
    # Compute cluster stats
    #
    # degree_sequence = sorted([d for n, d in cluster_graph.degree()], reverse=True)
    # degreeCount = Counter(degree_sequence)
    # degree_hist = pd.DataFrame(list(degreeCount.items()),
    #                           columns=['degree', 'count'])
    # degree_hist.set_index('degree', inplace=True)
    # degree_hist.sort_values('degree', inplace=True)
    # degree_hist.to_csv(histfilepath, sep='\t')
    nx.write_gml(cluster_graph, gmlfilepath)
    click_loguru.elapsed_time("final")
    return run_stats, cluster_graph, cluster_hist, any_hist, all_hist


def cluster_in_steps(seqfile, steps, min_id_freq=0, substrs=None, dups=None):
    """Cluster in steps from low to 100% identity."""
    try:
        inpath, dirpath = get_paths_from_file(seqfile)
    except FileNotFoundError:
        logger.error('Input file "%s" does not exist!', seqfile)
        sys.exit(1)
    stat_path = dirpath / (inpath.stem + STATFILE_SUFFIX)
    any_path = dirpath / (inpath.stem + ANYFILE_SUFFIX)
    all_path = dirpath / (inpath.stem + ALLFILE_SUFFIX)
    logsteps = [1.0] + list(
        1.0 - np.logspace(IDENT_LOG_MIN, IDENT_LOG_MAX, num=steps)
    )
    min_fmt = prettyprint_float(min(logsteps) * 100.0, 2)
    max_fmt = prettyprint_float(max(logsteps) * 100.0, 2)
    logger.info(
        f"Clustering at {steps} levels from {min_fmt}% to {max_fmt}% global"
        " sequence identity"
    )
    stat_list = []
    all_frames = []
    any_frames = []
    for id_level in logsteps:
        (
            stats,
            unused_graph,
            unused_hist,
            any_,
            all_,
        ) = homology_cluster(  # pylint: disable=unused-variable
            seqfile,
            id_level,
            min_id_freq=min_id_freq,
            substrs=substrs,
            dups=dups,
        )
        stat_list.append(stats)
        any_frames.append(any_)
        all_frames.append(all_)
    logger.info(f"Collating results on {seqfile}.")
    #
    # Concatenate and write stats
    #
    stats = pd.DataFrame(stat_list)
    stats.to_csv(stat_path, sep="\t")
    #
    # Concatenate any/all data
    #
    any_ = pd.concat(
        any_frames, axis=1, join="inner", sort=True, ignore_index=False
    )
    any_.to_csv(any_path, sep="\t")
    all_ = pd.concat(
        all_frames, axis=1, join="inner", sort=True, ignore_index=False
    )
    all_.to_csv(all_path, sep="\t")


def clusters_to_histograms(infile):
    """Compute histograms from cluster file."""
    try:
        inpath, dirpath = get_paths_from_file(infile)
    except FileNotFoundError:
        logger.error(f'Input file "{infile}" does not exist!')
        sys.exit(1)
    histfilepath = dirpath / f"{inpath.stem}-sizedist.tsv"
    clusters = pd.read_csv(dirpath / infile, sep="\t", index_col=0)
    cluster_counter = Counter()
    for unused_cluster_id, group in clusters.groupby(
        ["hom.cluster"]
    ):  # pylint: disable=unused-variable
        cluster_counter.update({len(group): 1})
    logger.info(f"Writing to {histfilepath}.")
    cluster_hist = pd.DataFrame(
        list(cluster_counter.items()), columns=["siz", "clusts"]
    )
    total_clusters = cluster_hist["clusts"].sum()
    cluster_hist["%clusts"] = cluster_hist["clusts"] * 100.0 / total_clusters
    cluster_hist["%genes"] = (
        cluster_hist["clusts"] * cluster_hist["siz"] * 100.0 / len(clusters)
    )
    cluster_hist.sort_values(["siz"], inplace=True)
    cluster_hist.set_index("siz", inplace=True)
    cluster_hist.to_csv(histfilepath, sep="\t", float_format="%06.3f")


def compare_clusters(file1, file2):
    """Compare cluster files."""
    path1 = Path(file1)
    path2 = Path(file2)
    commondir = Path(os.path.commonpath([path1, path2]))
    missing1 = commondir / "notin1.tsv"
    missing2 = commondir / "notin2.tsv"
    clusters1 = pd.read_csv(path1, sep="\t", index_col=0)
    print(f"{len(clusters1):d} members in {file1}")
    clusters2 = pd.read_csv(path2, sep="\t", index_col=0)
    print(f"{len(clusters2):d} members in {file2}")
    ids1 = set(clusters1["id"])
    ids2 = set(clusters2["id"])
    notin1 = pd.DataFrame(ids2.difference(ids1), columns=["id"])
    notin1.sort_values("id", inplace=True)
    notin1.to_csv(missing1, sep="\t")
    notin2 = pd.DataFrame(ids1.difference(ids2), columns=["id"])
    notin2.sort_values("id", inplace=True)
    notin2.to_csv(missing2, sep="\t")
    print("%d ids not in ids1" % len(notin1))
    print("%d ids not in ids2" % len(notin2))
    print(f"{len(clusters1):d} in {file1} after dropping")


def cleanup_fasta(
    set_path,
    fasta_path_or_handle,
    filestem,
    write_fasta=True,
    write_stats=True,
):
    """Sanitize and characterize protein FASTA files."""
    out_sequences = []
    ids = []
    lengths = []
    n_ambigs = []
    m_starts = []
    no_stops = []
    seqs = []
    sanitizer = Sanitizer(filestem)
    if hasattr(fasta_path_or_handle, "read"):
        handle = fasta_path_or_handle
    else:
        handle = fasta_path_or_handle.open("rU")
    with handle:
        for record in SeqIO.parse(handle, SEQ_FILE_TYPE):
            seq = record.seq.upper().tomutable()
            try:
                seq, m_start, no_stop, n_ambig = sanitizer.sanitize(seq)
            except ValueError:  # zero-length sequence after sanitizing
                continue
            seqs.append(str(seq))
            if write_fasta:
                record.seq = seq.toseq()
                out_sequences.append(record)
            ids.append(record.id)
            lengths.append(len(seq))
            n_ambigs.append(n_ambig)
            m_starts.append(m_start)
            no_stops.append(no_stop)
    if write_fasta:
        with (set_path / f"{filestem}.fa").open("w") as output_handle:
            SeqIO.write(out_sequences, output_handle, SEQ_FILE_TYPE)
    properties_frame = pd.DataFrame(
        list(zip(ids, lengths, n_ambigs, m_starts, no_stops, seqs)),
        columns=[
            "ID",
            "prot.len",
            "prot.n_ambig",
            "prot.m_start",
            "prot.no_stop",
            "prot.seq",
        ],
    )
    properties_frame = properties_frame.set_index(["ID"])
    if write_stats:
        propfile_path = set_path / protein_properties_filename(filestem)
        properties_frame.to_csv(propfile_path, sep="\t")
        frame_return = len(properties_frame)
    else:
        propfile_path = None
        frame_return = properties_frame
    return (
        filestem,
        propfile_path,
        frame_return,
        sanitizer.file_stats(),
    )


def compute_subclusters(cluster, cluster_size_dict=None):
    """Compute dictionary of per-subcluster stats."""
    subcl_frame = pd.DataFrame(
        [
            {
                "homo_id": i,
                "mean_len": g["len"].mean(),
                "std": g["len"].std(),
                "sub_siz": len(g),
            }
            for i, g in cluster.groupby("link")
        ]
    )
    subcl_frame["link"] = subcl_frame["homo_id"]
    subcl_frame["cont"] = subcl_frame["sub_siz"] == [
        cluster_size_dict[cluster_id] for cluster_id in subcl_frame["homo_id"]
    ]
    subcl_frame.loc[subcl_frame["cont"], "link"] = pd.NA
    subcl_frame.sort_values(
        ["sub_siz", "std", "mean_len"],
        ascending=[False, True, False],
        inplace=True,
    )
    subcl_frame["sub"] = list(range(len(subcl_frame)))
    subcl_frame.index = list(range(len(subcl_frame)))
    # normalized length is NaN for first element
    subcl_frame["norm"] = [pd.NA] + list(
        subcl_frame["mean_len"][1:] / subcl_frame["mean_len"][0]
    )
    subcl_dict = subcl_frame.set_index("homo_id").to_dict("index")
    #
    # subcluster attributes are done, now copy them into the cluster frame
    #
    for attr in ["norm", "std", "sub_siz", "sub", "link"]:
        cluster[attr] = [
            subcl_dict[cluster_id][attr] for cluster_id in cluster["link"]
        ]
    if subcl_frame["cont"].all():
        cluster["cont"] = 1
    else:
        cluster["cont"] = 0
    return cluster
